/*
 * Copyright 2024 KylinSoft Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include "baidunlpengine_p.h"
#include "servererror.h"
#include "token.h"
#include "util.h"

BaiduNlpEnginePrivate::BaiduNlpEnginePrivate() {}

BaiduNlpEnginePrivate::~BaiduNlpEnginePrivate() {
    clearContext();
    streamChatData_.clear();
}

std::string BaiduNlpEnginePrivate::engineName() { return "baidu"; }

void BaiduNlpEnginePrivate::setAppId(const std::string &appId) {
    appId_ = appId;
}

void BaiduNlpEnginePrivate::setApiKey(const std::string &apiKey) {
    apiKey_ = apiKey;
}

void BaiduNlpEnginePrivate::setSecretKey(const std::string &secretKey) {
    secretKey_ = secretKey;
}

std::string &BaiduNlpEnginePrivate::getAccessToken_() {
    if (accessToken_.empty()) {
        generateAccessToken_();
    }
    return accessToken_;
}

void BaiduNlpEnginePrivate::generateAccessToken_() {
    bool hasNetError = false;
    accessToken_ = token::getBaiduToken(apiKey_, secretKey_, hasNetError,
                                        lastErrorMessage_);
    if (accessToken_.empty()) {
        hasNetError
            ? resultCode_ = (int)ai_engine::ErrorCode::NetError
            : resultCode_ = (int)ai_engine::ErrorCode::AuthenticationFailed;
        fprintf(stderr, "Get baidu llm token failed.\n");
        return;
    }
    resultCode_ = (int)ai_engine::ErrorCode::Success;
}

void BaiduNlpEnginePrivate::setChatResultCallback(
    ai_engine::lm::nlp::ChatResultCallback callback) {
    chatResultCallback_ = callback;
}

void BaiduNlpEnginePrivate::setContextSize(int size) { contextSize_ = size; }

void BaiduNlpEnginePrivate::clearContext() { context_.clear(); }

namespace {
std::vector<std::string> &splitDataString(const std::string &dataStr) {
    static std::vector<std::string> datas;
    datas.clear();

    int oldIndex = 0;
    int newIndex = 0;
    while ((newIndex = dataStr.find_first_of("\n\n", newIndex)) !=
           std::string::npos) {
        datas.push_back(dataStr.substr(oldIndex, newIndex - oldIndex));
        newIndex += 2;
        oldIndex = newIndex;
    }
    return datas;
}
} // namespace

bool writeChatData(std::string data, intptr_t userData) {
    BaiduNlpEnginePrivate *engine = (BaiduNlpEnginePrivate *)userData;
    ai_engine::lm::nlp::ChatResultCallback callback =
        engine->chatResultCallback_;

    if (engine->isStopped_) {
        return false;
    }

    Json::Value resultJson = util::formatJsonFromString(data);
    if (token::isBaiduTokenExpiredByResult(resultJson)) {
        engine->needRefreshAccessToken_ = true;
        return false;
    }

    if (int errorCode = server_error::parseErrorCode(data)) {
        engine->lastErrorMessage_ = data;
        engine->resultCode_ = errorCode;
        fprintf(stderr, "Get baidu llm result failed %s.\n", data.c_str());
        return false;
    }

    std::string resultString = engine->tempStreamChatData_.data() + data;
    if (resultString.substr(resultString.length() - 2) != "\n\n") {
        engine->tempStreamChatData_ += data;
        return true;
    }
    engine->tempStreamChatData_.clear();

    std::vector<std::string> dataStrings = splitDataString(resultString);
    engine->resultCode_ = (int)ai_engine::ErrorCode::Success;
    for (std::string data : dataStrings) {
        data = data.substr(6, data.length() - 1);
        Json::Value obj = util::formatJsonFromString(data);
        Json::Value callbackObj;
        callbackObj["sentence_id"] = obj["sentence_id"];
        callbackObj["is_end"] = obj["is_end"];
        callbackObj["result"] = obj["result"];
        engine->streamChatData_.push_back(obj["result"]);
        callback(callbackObj.toStyledString());
    }
    return true;
}

void BaiduNlpEnginePrivate::addContext_(const std::string &role,
                                        const std::string &message) {
    int size = context_["messages"].size();
    context_["messages"][size]["role"] = role;
    context_["messages"][size]["content"] = message;
}

void BaiduNlpEnginePrivate::removeLastContext_() {
    int size = context_["messages"].size();
    if (size - 1 >= 0 &&
        context_["messages"][size - 1]["role"].asString() == "user") {
        Json::Value removeValue;
        context_["messages"].removeIndex(size - 1, &removeValue);
        return;
    }
    if ((size - 1) >= 0 && (size - 2) >= 0 &&
        context_["messages"][size - 1]["role"].asString() == "assistant" &&
        context_["messages"][size - 2]["role"].asString() == "user") {
        Json::Value removeValue;
        context_["messages"].removeIndex(size - 1, &removeValue);
        context_["messages"].removeIndex(size - 2, &removeValue);
        return;
    }
}

ai_engine::ErrorCode BaiduNlpEnginePrivate::chat(const std::string &message) {
    isStopped_ = false;
    if (contextSize_ == 0) {
        clearContext();
    }
    addContext_("user", message);
    context_["stream"] = true;
    std::string postString = context_.toStyledString();

    cpr::Url url{"https://aip.baidubce.com/rpc/2.0/ai_custom/v1/"
                 "wenxinworkshop/chat/completions_pro"};
    cpr::Parameters parameters{
        cpr::Parameters{{"access_token", getAccessToken_()}}};
    cpr::Header header{cpr::Header{{"Content-Type", "application/json"}}};
    cpr::Body body{cpr::Body{{postString}}};
    cpr::Response response =
        cpr::Post(url, parameters, header, body,
                  cpr::WriteCallback{writeChatData, (intptr_t)this});

    if (!processChatResponse_(response)) {
        return chat(message);
    } else {
        return (ai_engine::ErrorCode(resultCode_));
    }
}

bool BaiduNlpEnginePrivate::processChatResponse_(cpr::Response &response) {
    if (needRefreshAccessToken_) {
        removeLastContext_();
        streamChatData_.clear();
        generateAccessToken_();
        needRefreshAccessToken_ = false;
        return false;
    }

    if (isStopped_) {
        removeLastContext_();
        streamChatData_.clear();
        return true;
    }

    if (response.error.code == cpr::ErrorCode::REQUEST_CANCELLED &&
        resultCode_ > 0) {
        removeLastContext_();
        streamChatData_.clear();
        resultCode_ = (int)server_error::errorCode2nlpResult(resultCode_);
        return true;
    }

    if (response.error) {
        removeLastContext_();
        streamChatData_.clear();
        lastErrorMessage_ = response.error.message;
        resultCode_ = (int)ai_engine::ErrorCode::NetError;
        fprintf(stderr, "Net error: %s.\n", response.error.message.c_str());
        return true;
    }

    std::string assistantcontent;
    for (Json::Value &value : streamChatData_) {
        assistantcontent.append(value.asString());
    }
    addContext_("assistant", assistantcontent);
    context_.removeMember("stream");
    streamChatData_.clear();
    resultCode_ = (int)ai_engine::ErrorCode::Success;
    return true;
}

void BaiduNlpEnginePrivate::stopChat() { isStopped_ = true; }

ai_engine::ErrorCode BaiduNlpEnginePrivate::initChatModule() {}

ai_engine::ErrorCode BaiduNlpEnginePrivate::destroyChatModule() {}
